from minitap.mobile_use.sdk.types.agent import ApiBaseUrl


DEFAULT_PROFILE_NAME = "default"
DEFAULT_HW_BRIDGE_BASE_URL = ApiBaseUrl(
    scheme="http",
    host="localhost",
    port=9999,
)
DEFAULT_SCREEN_API_BASE_URL = ApiBaseUrl(
    scheme="http",
    host="localhost",
    port=9998,
)
